/*
 * extension.js
 * Gnome3 Arduino Control Extension
 *
 * Simple bidirectional control of Arduino outputs through a gnome-shell panel menu
 *
 *
 * Author: Simon Junga (simonthechipmunk@gmx.de)
 *
 * This extension is mostly based on code found in the following gnome-extensions:
 * shutdowntimer@lusk.cz
 * ShutdownTimer@neumann
 * AdvancedVolumeMixer@harry.karvonen.gmail.com
 * chatstatus@zeten30.gmail.com
 * backslide@codeisland.org
 * lockkeys@vaina.lt
 *
 * Thanks to the authors of these great extensions. Since there is very little information on writing Gnome3 extensions,
 * well documented ones are a blessing.
 *
 * This programm is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 */


//***// imports:

// icons and labels
const St = imports.gi.St;

// main functionality
const Main = imports.ui.main;

// menu items
const Panel = imports.ui.panel;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;
const Slider = imports.ui.slider;

// utilities for external programs and command line
const Util = imports.misc.util;
const GLib = imports.gi.GLib;
const Gio = imports.gi.Gio;

// messagetray notifications
const MessageTray = imports.ui.messageTray;

// clutter
const Clutter = imports.gi.Clutter;

// own imports
const Me = imports.misc.extensionUtils.getCurrentExtension();
const Prefs = Me.imports.prefs;




// define global variables
let device, panelmenu, sockClient, sockConnection, output_reader, cancel_read, Switch1, Slider1, Sensor1;
let Button1, inStr;


const Switch1Actor = "out2";
const Slider1Actor = "out3";
const Sensor1Actor = "sense0";




//***// basic extension functions

function init() {
	// initialize preferences
	Prefs.init();

}



function enable() {
	// set up the client and Gcancellable oeration for the async call 
	sockClient = new Gio.SocketClient();
	cancel_read = new Gio.Cancellable();


	// connect to socket server
	try {
		sockConnection = sockClient.connect_to_host(Prefs._getIP() + ":" + Prefs._getPort(), null, null);
	}

	catch(err) {
		_CallNotify("Arduino Control", "Error connecting to Socket Server. Please check it's IP Address in the Extension Settings and whether the server is set up properly." 
			+ "\n" + "\n" + err, false);
		return;
	}

	// call function to create panelmenu
	_ArduinoPanelMenu(true);


	// create new input stream from sockConnection
	inStr = new Gio.DataInputStream({ base_stream: sockConnection.get_input_stream() });
	// create new consistent stream for output_reader from inStr (otherwise Gnome-Shell will crash on read_line_async when the inputstream gets closed by the server)
	output_reader = new Gio.DataInputStream({ base_stream: inStr });
	// start waiting for server messages and initialize async read on firstrun
	output_reader.read_line_async(0, cancel_read, _SocketRead, null);


	// establish initial connection to the server by sending some randomness (needed for Arduino with Ethernet Module)
	_SocketSend("client", "connect");


}




function disable() {

	// cancel the async read call
	cancel_read.cancel();

	try {
	// call function to remove panelmenu
	_ArduinoPanelMenu(false);

	// close socket connection
	sockConnection.close(null);

	}

	catch(err) {
	// nothing to do
	}


}






//***// extension functions

function _ArduinoPanelMenu(set) {
// create the extension menu in the status area with slider and toggle switch

  if (set) {

	// create slider with default value 0
	Slider1 = new PopupMenu.PopupMenuItem("");
	let Slider1Icon = new St.Icon({ icon_name: 'keyboard-brightness-symbolic', style_class: 'popup-menu-icon' });
	Slider1.actor.add(Slider1Icon);
	Slider1Item = new Slider.Slider(0);
	Slider1Item.connect('value-changed', _Slider1Change);
	Slider1.actor.add(Slider1Item.actor, { expand: true });


	// create toggle switch
	Switch1 = new PopupMenu.PopupSwitchMenuItem("");
	Switch1.label.text = 'Light';
	Switch1.connect('toggled', _Switch1Toggle);


	// create menu item for sensor values
	Sensor1 = new PopupMenu.PopupMenuItem("", { reactive: false });
	let Sensor1Icon = new St.Icon({ icon_name: 'media-playlist-consecutive-symbolic', y_align: Clutter.ActorAlign.START, style_class: 'popup-menu-icon' });
	Sensor1.label.text = 'Sensor ';
	Sensor1Item = new St.Label({ text: "--", style_class: "sm-label" });
	Sensor1.actor.add(Sensor1Icon);
	Sensor1.actor.add(Sensor1Item);


	// create seperate panelmenu button on the top panel
	panelmenu = new PanelMenu.Button(0.0);
	let icon = new St.Icon({icon_name: 'dialog-information-symbolic', style_class: 'system-status-icon'});
	panelmenu.actor.add_actor(icon);
	// add Switch1, Slider1 and Sensor1 to it
	panelmenu.menu.addMenuItem(Switch1);
	panelmenu.menu.addMenuItem(Slider1);
	panelmenu.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
	panelmenu.menu.addMenuItem(Sensor1);


	// add panelmenu to the main panel
	Main.panel.addToStatusArea("ArduinoControl", panelmenu, 0, "right");

	}


  else {
	// remove the panelmenu
	panelmenu.destroy();

	}

}





function _Switch1Toggle() {
// react on switch toggle

	if(Switch1.state) {
		_SocketSend(Switch1Actor, 1);
	} 

	else {
		_SocketSend(Switch1Actor, 0)
	}

}




function _Slider1Change() {
// react on slider position change
	let sliderval = Math.floor(Slider1Item.value * 255);
	_SocketSend(Slider1Actor, sliderval);


}





function _CallNotify(header, text, transient) {
// Call System Notification

	// create notification source
	let source = new MessageTray.Source("ArduinoControl", 'media-optical-cd-audio-symbolic');
	let notification = new MessageTray.Notification(source, "" + header, "" + text);
	// display the notification
	Main.messageTray.add(source);
	source.notify(notification);
	// set the notification style
	notification.setTransient(transient);
}





function _SocketSend(actor, value) {
// Send data to server

	let outStr = sockConnection.get_output_stream();
	outStr.write(actor + ':' + value + '\n', null);

}





function _SocketRead(gobject, async_res, user_data) {
// Callback for reading incoming Server messages

	var splitarray;

	// finish async_read call
	let [lineout, charlength, error] = gobject.read_line_finish(async_res);

	// start new read_async call
	cancel_read.reset();
	output_reader.read_line_async(0, cancel_read, _SocketRead, null);

	// process received data
	_ProcessData(lineout);


}





function _ProcessData(data) {
// Read Server stream and process it

	var splitarray;
	// split received data into actor and value
	splitarray = ("" + data).split(":");

	let switch_set = parseInt(splitarray[1]);
	let slider_set = switch_set / 255;

	if ((splitarray[0] == Switch1Actor) && (Switch1.state != switch_set)) {
		// handle Switch1
		Switch1.setToggleState(switch_set);
		}

	else if ((splitarray[0] == Slider1Actor) && (Slider1Item.value != slider_set)) {
		// handle Slider1
		Slider1Item.setValue(slider_set);
	}
	
	else if (splitarray[0] == Sensor1Actor) {
		// update Sensor Data in panelmenu
		Sensor1Item.text = splitarray[1];
	}

 
	else if( splitarray[0] == "error") {

		_CallNotify("Arduino Control", "An Error Code was received from the Server:" + "\n" + data, true);
	}


} 


